use crate::cmd::config::app_config;
use crate::ems::{Consumer, Producer, Service};
use crate::internal::modbus::code::{ModbusCode, ModbusReqCode, RegisterType, Rw};
use crate::internal::modbus::{read_csv_to_code, slice_sequential, MAX_BIT_CNT, MAX_WORD_CNT};
use crate::internal::utils;
use crate::internal::utils::mqtt::mqtt_client;
use crate::{connect_or_retry, publish_data, read_registers_tcp};
use anyhow::{bail, Error};
use async_trait::async_trait;
use log::{error, info};
use mosquitto_rs::QoS;
use std::net::SocketAddr;
use std::sync::Arc;
use std::time::Duration;
use tokio::sync::{Mutex, MutexGuard};
use tokio::time::Instant;
use tokio_modbus::client::{Reader, Writer};

///
/// 高特BMS
/// Modbus-TCP
pub struct GoldBms {
    pub id: String,
    pub _producer: Producer,
    pub consumer: Arc<Mutex<Consumer>>,
    pub modbus_code: Arc<Mutex<Vec<ModbusCode>>>,
    pub ctx: Arc<Mutex<tokio_modbus::client::Context>>,
}

impl GoldBms {
    pub async fn new(producer: Producer, consumer: Consumer) -> anyhow::Result<Self> {
        let id = utils::generate_random_str(12);
        utils::log::init_log("inpower_iot_mgc_rs::ems::bms::*", "bms/bms.log").await;
        info!("开始初始化BMS[{}]...", id);
        let config = read_csv_to_code("./config/gold_bms.csv").map_err(|e| {
            error!("协议生成失败:{}", e);
            anyhow::anyhow!(e)
        })?;
        info!("成功加载BMS协议配置，共{}个寄存器", config.len());
        let ctx = connect_modbus_tcp().await?;
        info!("{}初始化成功!", id);
        Ok(GoldBms {
            id,
            _producer: producer,
            consumer: Arc::new(Mutex::new(consumer)),
            modbus_code: Arc::new(Mutex::new(config)),
            ctx: Arc::new(Mutex::new(ctx)),
        })
    }

    //读取输入寄存器
    async fn read_input_register(&self, mut vec: Vec<&mut ModbusCode>) -> anyhow::Result<()> {
        read_registers_tcp!(self, vec, MAX_WORD_CNT, read_input_registers, |x| x)
    }

    //读取离散输入寄存器
    async fn read_discrete_input(&self, mut vec: Vec<&mut ModbusCode>) -> anyhow::Result<()> {
        read_registers_tcp!(self, vec, MAX_BIT_CNT, read_discrete_inputs, |x: bool| x
            as u16)
    }

    //读取保持寄存器
    async fn read_holding_register(&self, mut vec: Vec<&mut ModbusCode>) -> anyhow::Result<()> {
        read_registers_tcp!(self, vec, MAX_WORD_CNT, read_holding_registers, |x| x)
    }

    //推送消息
    async fn push(&self, vec: &[ModbusCode], time: Instant) -> anyhow::Result<()> {
        let client = &mqtt_client().await?.mosq;
        // 发布遥测数据 (YC)
        publish_data!(client, vec, "/pds/bms/0/yc", RegisterType::InputRegister);
        // 发布遥信数据 (YX)
        publish_data!(client, vec, "/pds/bms/0/yx", RegisterType::DiscreteInput);
        // 发布遥调数据 (YT)
        publish_data!(client, vec, "/pds/bms/0/yt", RegisterType::HoldingRegister);
        info!("MQTT ↑↑↑, took {:?}", time.elapsed());
        Ok(())
    }
}

#[async_trait]
impl Service for GoldBms {
    async fn south(&self) {
        let code_clone = self.modbus_code.clone();
        loop {
            tokio::time::sleep(Duration::from_millis(app_config().emu.bms.interval)).await;
            let mut modbus_guard = code_clone.lock().await;
            let time = Instant::now();
            if let Err(e) = self
                .read_discrete_input(
                    modbus_guard
                        .iter_mut()
                        .filter(|it| it.code == RegisterType::DiscreteInput)
                        .collect(),
                )
                .await
            {
                handle_modbus_error(&e, &self.ctx, "读取离散寄存器失败").await;
            }

            if let Err(e) = self
                .read_input_register(
                    modbus_guard
                        .iter_mut()
                        .filter(|it| it.code == RegisterType::InputRegister)
                        .collect(),
                )
                .await
            {
                handle_modbus_error(&e, &self.ctx, "读取输入寄存器失败").await;
            }

            if let Err(e) = self
                .read_holding_register(
                    modbus_guard
                        .iter_mut()
                        .filter(|it| {
                            it.code == RegisterType::HoldingRegister && it.rw == Some(Rw::RW)
                        })
                        .collect(),
                )
                .await
            {
                handle_modbus_error(&e, &self.ctx, "读取保持寄存器失败").await;
            }
            if let Err(e) = &self.push(modbus_guard.as_slice(), time).await {
                error!("MQTT ↑↑↑ error: {}", e);
            }
        }
    }

    async fn north(&self) {
        let consumer = self.consumer.clone();
        loop {
            let ctx = self.ctx.clone();
            let modbus_code_clone = self.modbus_code.clone();
            if let Ok(msg) = consumer.lock().await.recv().await {
                if msg.id == self.id {
                    info!("MQTT ↓↓↓: {}", msg);
                    let instant = Instant::now();
                    let modbus_code = modbus_code_clone.lock().await;
                    let mut req_code = ModbusReqCode::build(modbus_code.as_slice(), &msg);
                    let vec = slice_sequential(req_code.as_mut_slice(), MAX_WORD_CNT);
                    let context_guard = ctx.lock().await;
                    match down_link(vec, context_guard).await {
                        Ok(()) => info!("modbus ↓↓↓, took {:?}", instant.elapsed()),
                        Err(e) => {
                            error!("modbus ↓↓↓ error:{}", e)
                        }
                    }
                }
            }
        }
    }
}

async fn connect_modbus_tcp() -> anyhow::Result<tokio_modbus::client::Context> {
    let config = &app_config().emu.bms.gold_bms;
    let addr: SocketAddr = format!("{}:{}", config.host, config.port).parse()?;
    connect_or_retry!(tokio_modbus::client::tcp::connect(addr), "[BMS] ModbusTCP")
}

async fn handle_modbus_error(e: &Error, ctx: &Mutex<tokio_modbus::client::Context>, log_str: &str) {
    //错误是tokio_modbus的则重新连接
    if let Some(_) = e.downcast_ref::<tokio_modbus::Error>() {
        let mut ctx = ctx.lock().await;
        if let Ok(context) = connect_modbus_tcp().await {
            *ctx = context;
        }
    } else {
        error!("{log_str}: {}", e);
    }
}

async fn down_link(
    req_codes: Vec<&mut [ModbusReqCode<'_>]>,
    mut context_guard: MutexGuard<'_, tokio_modbus::client::Context>,
) -> anyhow::Result<()> {
    for codes in req_codes.into_iter() {
        if codes.len() > 1 {
            let words: Vec<_> = codes.iter().map(|it| it.trans_val()).collect();
            context_guard
                .write_multiple_registers(codes[0].addr, words.as_slice())
                .await??;
        } else {
            context_guard
                .write_single_register(codes[0].addr, codes[0].trans_val())
                .await??
        }
    }
    Ok(())
}
